/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
#define _GNU_SOURCE
#include <stdio.h>

#include <stdlib.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "atsend.h"
#include "cmds.h"
#include "common.h"
#include "parse.h"

extern void closePort(int final);

extern int  	    serverMode;
extern char 	    callerId[MAXLEN];
extern int  	    repeatNow;
extern type_key*    repeatCmdPtr;

static void 	  updateTimerLastData (timerCmd* timer, char *data);
static timerCmd*  createTimer(cmdItem *ci);
static int        processTimerCmd(cmdItem* ci);

char tmptext[MAXCMDLEN];
int  remoteOn       = 1;

timerCmd* timers  = NULL;

// This is not clean-up function. It just finished the connection.
static void finishAll()
{
        // finish events forwarding
        printf("Finish events forwarding\n");
        logger("INFO","Finish events forwarding");

        if (serverMode == 0) {
	
		sendCMER(CMER_OFF);

                //handle_command(answer); // Does it needed ?
                // and resend No back
                //sendSyncCommand("AT+CKPD=\"e\"\r", answer, sizeof(answer), NULL);    // Does it needed ?
        }
        closePort(1);

        return;
}

static int execInFork(char *cmd)
{
	int ret;
        int cpid = fork();
        if( cpid < 0) {
        	logger("ERROR", "Can not fork!");
                return -1;
        }

        if(cpid) {      // father
        	waitpid(cpid,&ret,0);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
        } else {        // child
        	logger("INFO", "Close port in child");
                closePort(0);
                 
                ret = system(cmd);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
                exit(ret);
        }

        return ret;
}

//static char* fBuffer;

//static void freeResultsBuffer() 
//{
//	free(fBuffer);
//	fBuffer = NULL;
//}

static char *readResultsFile() 
{
	FILE *fp;
	struct stat buf;
	
	if(stat(TOFILE, &buf) == -1) {
		return NULL;
	}
	
	long fLen = buf.st_size;
	sprintf(tmp, "readResultsFile >%ld<", fLen);
	logger("DEBUG",tmp);
	
	fp=fopen(TOFILE,"r");
	if (fp == NULL) {
		return NULL;
	}
    
    	char * fBuffer = NULL;
    	fBuffer = calloc(fLen+1,1);
        if (fBuffer == NULL) {
        	logger("ERROR","no more memory!");
		return NULL;
        }
    
	fread(fBuffer, sizeof(char), fLen, fp);
    
	//logger("DEBUG","readResultsFile: exiting");
	return fBuffer;
}

static char* substParams(char* in, cmdParams *params) {


        if (in == NULL) {
                logger("ERROR","substParams(): null input");
                return NULL;
        }
        sprintf(tmp, "substParams >%s<", in);
        logger("INFO",tmp);

        if (*in == '\0') { // Empty command. Skip it.
            return NULL;
        }
        
        char *command =  strdup(in);
        char *ptr     = command;
        char *papameterized = NULL;

        int pc = strlen(PARAM_CALLID);
        int pt = strlen(PARAM_TIME);
        int pm = strlen(PARAM_MODE);
        int pp = strlen(PARAM_PARAM);
        int pi = strlen(PARAM_INDEX);
        int pd = strlen(PARAM_CFGDIR);
        
        // calculate size to allocate
        int needSize = strlen(ptr)+1; 
        int nParams  = 0;
        while(ptr != NULL) {
                papameterized = strstr(ptr, PARAM_START);   // $(
                if (papameterized == NULL) { // Just copy
                        break;
                } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
                	needSize += strlen(callerId) - pc - 2;
                        ptr = papameterized + pc + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
               		needSize += 16 - pt - 2;
                        ptr = papameterized + pt + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
                	char *m = getMode();
                        needSize += strlen(m) - pm - 2;	
                        ptr = papameterized + pm + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
                        if (params != NULL && params->value != NULL) {   // value can be empty
                        	needSize += strlen(params->value) - pp - 2;
                	}
                        ptr = papameterized + pp + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
                        if (params != NULL && params->index[0] != '\0') {
                        	needSize += strlen(params->index) - pi - 2;
                        }
                        ptr = papameterized + pi + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {
			int c = 0;
                        char *d;
                        if ((d = getCfgDir()) != NULL) {
				c = strlen(d);
                                needSize += c - pd - 2;
                        }
                        ptr = papameterized + pd + 2;
                        nParams++;
                } else {
                	break;
        	}
        }
        
        if (nParams == 0) {		// non-parametrized command
        	return command; 
        }

        papameterized = NULL;
        ptr = command;
        
        sprintf(tmp, "Allocate >%d< bytes for parametrized command", needSize);
        logger("DEBUG",tmp);

	char* cmd  = (char*) calloc(needSize+1,1);
       
        while(command != NULL) {
        	sprintf(tmp, "cur cmd >%s<", command);
                logger("DEBUG",tmp);

                papameterized = strstr(command, PARAM_START);   // $(
        
                if (papameterized != NULL) {
                	sprintf(tmp, "papameterized >%s<", papameterized);
                        logger("DEBUG",tmp);
                }
        
                if (papameterized == NULL) { // Just copy
                	logger("DEBUG", "no more parameters");
                        strcat(cmd, command);
                        break;
                } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
                	
                        *papameterized = '\0';
                        
                        strcat(cmd, command);		// First part
                        strcat(cmd, callerId);		// Add params
                
                        command = papameterized + pc + 2;
                    
                } else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
                	
                        *papameterized = '\0';

                        strcat(cmd, command);		// First part
                        time_t logtime;			// Add time
                        time(&logtime);

                        struct tm curtime;
                        localtime_r(&logtime, &curtime);

                        char buf[16];
                        sprintf(buf,"%d:%02d", curtime.tm_hour, curtime.tm_min);

                        strcat(cmd, buf);

                        command = papameterized + pt + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
                	
                        *papameterized = '\0';
                       
                        strcat(cmd, command); 		// First part
                        strcat(cmd, getMode());		// Add params
                                
                        command = papameterized + pm + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
                	// This is dirty trick to make Bemused commands like VOLM<volume> works
                        // also it is used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL) {   // value can be empty
                        	logger("WARNING", "No parameters supplied!");

                        	strcat(cmd, PARAM_START);
                        	strcat(cmd, PARAM_PARAM);
                        } else {
                        	strcat(cmd, command);		// First part
                        	strcat(cmd, params->value);	// Add params
                	}
                        command = papameterized + pp + 2;

                } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
                	// Used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL || params->index[0] == '\0') {
                        	logger("WARNING", "No parameters supplied!");

                        	strcat(cmd, PARAM_START);
                        	strcat(cmd, PARAM_PARAM);
                        } else {
                        	strcat(cmd, command);		// First part
                        	strcat(cmd, params->index);	// Add params
                        }
                        command = papameterized + pi + 2;

                } else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {
                
                        *papameterized = '\0';
                
                        // First part
                        strcat(cmd, command);

                        // Add cfg. dir.
                        char *d;
                        if ((d = getCfgDir()) != NULL) {
				strcat(cmd, d);
                        }
                
                        command = papameterized + pd + 2;
                
                } else {    // Just copy
                	strcat(cmd, command);
                        break;
                }
        }
        int cl = strlen(cmd);
        sprintf(tmp, "final cmd >%s< %d", cmd,cl);
        logger("DEBUG",tmp);
        
        free(ptr);
        
        return cmd;
}

static int execSimpleCmd(char *cmd, int toFile)
{
	if (cmd == NULL) {
        	logger("DEBUG", "execSimpleCmd null input");
        	return EXIT_OK;
        }
        sprintf(tmp, "execSimpleCmd >%s<", cmd);
        logger("INFO",tmp);

        if (remoteOn != 1) {
                logger("DEBUG", "Remote if OFF. Skip event");
                return EXIT_OK;
        }

        int add = 0;
        if (toFile) { 
        	add = strlen(TOFILE) + 4;
        } else {
        	add = 2;
        }
	char* p = calloc(strlen(cmd) + add,1);
        if (p == NULL) {
        	return EXIT_NOK;
        }
        strcpy(p,cmd);
         
        if (toFile) {       // We needs to wait results, so do not run in background
        	strcat(p, " > ");		
                strcat(p, TOFILE);
        } else {
        	if (! index(p, '&')) {
                	strcat(p, "&");
                }
        }
        int ret = execInFork(p);
        free(p);
        
        return (ret == -1 ? EXIT_NOK : EXIT_OK);
}


static int processSendCmd(cmdItem* ci)
{
	int ret = EXIT_OK;

        char all[MAXMAXLEN];
        strncpy(all, ci->descr, MAXMAXLEN-1);

        char *tag  = strtok(all, ",");
        if (tag == NULL) {
        	logger("DEBUG", "processSendCmd(): incorrect input (1)");                
        	return EXIT_NOK;
        }
        if (strlen(ci->descr) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        	logger("DEBUG", "processSendCmd(): incorrect input (2)");                
                return EXIT_NOK;
        }
        char *ptr   = tag + strlen(tag) - 1;
        // strip spaces from tail
        while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
        	*ptr = '\0';
        	ptr--;
        }
        ptr = tag + strlen(tag) + 1;
        // strip spaces from head of value
        while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        	ptr++;
        }

        sprintf(tmp, "processSendCmd() is >%s< >%s<",tag,ptr);
        logger("DEBUG", tmp);

        if (strcmp(tag, SEND_STRING) == 0) {
        	ret = writeSocketConn(ptr);
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeSocketConn()");
                }
        } else if (strcmp(tag, SEND_BYTES) == 0) {
        	ret = writeBytesSocketConn(ptr);
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeBytesSocketConn()");
                }    
        } else {
        	logger("DEBUG", "processSendCmd(): incorrect input (3)");                
                return EXIT_NOK;
        }
        return ret;
}

static void sendData(int type, int bytesFlag, char *data) 
{
	int ret;
	if (type == ID_SEND && bytesFlag) {
		ret = writeBytesSocketConn(data);
	} else {
		ret = writeSocketConn(data);
	}
	if (ret != EXIT_OK) {
		logger("DEBUG","Fails in write(Bytes)SocketConn()");
	}
}

static int macroCmd(char *macro, cmdParams* params) 
{
	logger("DEBUG","macroCmd");
        
        if (macro == NULL) {
        	return EXIT_NOK;
        }
 	int ret;
	type_key *k = findItem(macro, &ret, NULL);

	if (k && ret == FLAG_EXACT) {
		ret = handleCmdByKey(k, params);
		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
			return ret;
		}
	}
	return EXIT_OK;
}

void execDynamically(char *command) 
{
	cmdItem* tmpCmd = NULL;
	storeCmds(&tmpCmd, command); // parse command on-the-fly 
                
	processCommands(tmpCmd,NULL);
                
	freeCmds(tmpCmd);
	return;
}

static int loadCmds(char *file) 
{
	logger("DEBUG","loadCmds");
        if (file == NULL) {
        	return EXIT_NOK;
        }

	FILE *fp;
        char *command = NULL;
        size_t len = 0;
        ssize_t read;
	
	fp=fopen(file,"r");
	if (fp == NULL) {
		return EXIT_NOK;
	}
        
	while ((read = getline(&command, &len, fp)) != -1) {
        	sprintf(tmp,"Retrieved line %s of length %d ", command, (int) read);
		logger("DEBUG",tmp);
                
		execDynamically(command);

        }
        if (command) {
        	free(command);
        }
	return EXIT_OK;
}

static int includeCmd(char *file) 
{
	logger("DEBUG","includeCmd");
        if (file == NULL) {
		return EXIT_NOK;
        }

        load_cfg(file, 0);
        
        //printKeys();
        printConf();
        
	sprintf(tmp,"Mode is %s", getMode());
	logger("DEBUG",tmp);
        
	return EXIT_OK;
}
 
static void changeReplace2Add(char *pi) 
{
	char *pi2 = pi + 7;     // strlen("replace")
        *pi = '\0';
        strcat(pi,"add");
        pi += 3; 	     	// strlen("add")

        while (*pi2 != '\0') {
        	*pi = *pi2;
                pi++;
                pi2++;
        }
        *pi = '\0'; 
}

static int execCmdAndGetResults(int type, cmdItem* ci, timerCmd* timer, cmdParams *params) 
{
	logger("DEBUG","execCmdAndGetResults");

        if (serverMode == SERVER_NO || remoteOn != 1 || ci->exec == NULL) {
        	return EXIT_OK;
        }
        
        char* sExec = substParams(ci->exec, params);
        if (sExec == NULL) {
        	logger("DEBUG", "Empty exec. part after substParams()");
                return EXIT_NOK;
        }

        int ret = execSimpleCmd(sExec,1);
        if (ret != EXIT_OK) {
                sprintf(tmp, "Return code from execSimpleCmd() is %d",ret);
        	logger("DEBUG", tmp);
                free(sExec);
                return ret;
        }

        char* sDescr = substParams(ci->descr, params);
        if (sDescr == NULL) {
        	logger("DEBUG", "Empty command after substParams()");
                return EXIT_NOK;
        }
        
        char *res  = readResultsFile();
        char *ptrRes = res;
        int resLen = strlen(res);
	logger("DEBUG", "Empty command after readResultsFile()");

        if (timer != NULL) {
        	if (timer->lastValue != NULL && strcmp(timer->lastValue,res) == 0) {
                	logger("DEBUG","No needs to send the same data.");
                	free(sDescr);
                	free(ptrRes);
                	return EXIT_OK;
                } else {
                	updateTimerLastData(timer, res);	// Update "last data"
                }
        }

        if (type == ID_SEND) {  // Just send string
        	
                int bytesFlag = 0;
                if (strstr(sDescr, SEND_BYTES)) {
                	bytesFlag = 1;
                }
                sendData(type, bytesFlag, res);
                
        } else {                // ID_SET
		logger("DEBUG", "ID_SET");

                // Replace all ");" inside buffer to "@@" since Java Client will separate commands by ");"
                char *p;
                while ((p = strstr(res,");")) != NULL) {
                        *p = '@';
                        p++;
                        *p = '@';
                }

                char chunk[256];    
                strcpy(chunk,CMD_SET);

                strcat(chunk,"(");  // For example, Set(title
                strcat(chunk,sDescr);
            	
                if (strstr(chunk,SET_TEXT) != NULL) {
                	// long commands
                        // Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
                        // if string longer than 256 characters) we have to separate result to several chunks
                
                        int doLoop  = 1;   
                	int maxChunks = getMaxTextSize();
                	strcat(chunk,",");

                        while (doLoop > 0) {
                
                                char* pi;
                                if (doLoop != 1 && (pi = strstr(chunk,"replace")) != NULL ) {
                                	changeReplace2Add(pi); 
                                }
                		int headLen = strlen(chunk); 
                                 
                                if (maxChunks != -1 && doLoop >= maxChunks) {     // send only about 240*maxChunks data
                                        strncat(chunk, res, 200 - headLen);
                                        res += (200 - headLen); 
                                        strcat(chunk,"\n!!! file too long to show it in full !!!");
                                        doLoop = -1;
                                } else if (strlen(res) + headLen > 253) {             // 253 + ");\0" = 256
                                        strncat(chunk, res, 253 - headLen);
                                        res += (253 - headLen);              	// and move pointer
                                        doLoop++;
                                } else {                                	// last portion
                                        strcat(chunk, res);
                                        doLoop = -1;
                                }
                
                                strcat(chunk,");");

                                //sprintf(tmp, "set-text %d >%s<", doLoop,chunk);
                                //logger("INFO",tmp);
                                
                                sendData(type, 0, chunk);
                                *(chunk + headLen) = '\0';
                                
                                // no needs since JC will show text in dynamic
                                //if (doLoop == -1) {	// send final replay
                                //        strcat(chunk, "<<);");
                                //        sendData(type, 0, chunk);
                                //        *(chunk + headLen) = '\0';
                                //} 
                        }
                
                        logger("INFO","set-text finished");
               } else if (strstr(chunk,SET_LIST) != NULL ||
                          strstr(chunk,SET_FMGR) != NULL || 
                          strstr(chunk,SET_MENU) != NULL ||
                          strstr(chunk,SET_UPLD) != NULL) {
                       // long commands
                       // Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
                       // if string longer than 256 characters) we have to separate result to several chunks

                       int doLoop = 1;        

                       char* onePart = strtok(res,",");
                       while (doLoop > 0) {
                               
                                char* pi;
                                if (doLoop != 1 && (pi = strstr(chunk,"replace")) != NULL ) {
                                	changeReplace2Add(pi); 
                                }    
                		int headLen = strlen(chunk); 
                                int cLen = headLen;
				if (doLoop == 1 && strstr(chunk,SET_UPLD) != NULL) {
					strcat(chunk, ",");
					strcat(chunk, onePart);
					headLen = strlen(chunk);
					cLen    = headLen;
					onePart = strtok(NULL,",");
				}
                       
                              	while (onePart != NULL && cLen + strlen(onePart) < 252) {
                                         strcat(chunk, ",");
                                         strcat(chunk, onePart);
                                         cLen = strlen(chunk);
                                         onePart = strtok(NULL,",");
                              	}
                       
                            	if (strlen(chunk) == headLen && resLen != 0) { // We fall into infinite loop
                                	strcat(chunk, ",");
                                        strcat(chunk, "String too long");
                                        onePart = strtok(NULL,",");
                            	}
                             	if (resLen == 0) {
                              		strcat(chunk, ",");
                             	}
                       
                             	if (onePart == NULL) {
                            		doLoop = -1;    // Last portion
                             	} else {
                                       doLoop++;
                            	}

                            	strcat(chunk,");");
                       
                        	sendData(type, 0, chunk);
                       		*(chunk + headLen) = '\0';
                                
                                if (doLoop == -1 && strstr(chunk,SET_UPLD) != NULL) {	// send final replay
                                        strcat(chunk, ",<<);");
                                        sendData(type, 0, chunk);
                                        *(chunk + headLen) = '\0';
                                } 
               		}
        	} else {    // short commands. Are all of them really short (<256 bytes?)
                	strcat(chunk,",");
                        strcat(chunk,res);
                        strcat(chunk,");");
		      	//printf("CHUNK: %s\n", chunk);
                        sendData(type, 0, chunk);
          	}
  	}
	//freeResultsBuffer();
        free(ptrRes);
	free(sDescr);

        logger("INFO","execCmdAndGetResults finished");
        return EXIT_OK;
}

static int processOneCommand(cmdItem* ci, cmdParams* p)
{
	 int ret;

         if (ci->exec == NULL || ci->descr == NULL) {
         	if (ci->descr == NULL) {
	         	sprintf(tmp, "processOneCommand>no description< >%d<", remoteOn);
                }        
         	if (ci->exec == NULL) {
	         	sprintf(tmp, "processOneCommand>no exec params< >%d<", remoteOn);
                }        
         } else {
         	sprintf(tmp, "processOneCommand >%s< >%s< >%d<", ci->descr, ci->exec, remoteOn);
         }
         logger("INFO",tmp);
         
         if (ci->type == ID_REMOTEON) {
                 logger("INFO", "Switch remote on");
                 remoteOn = 1;        
                 return EXIT_OK;
         }
         if (remoteOn != 1) {
         	logger("INFO", "Remote is off. Skip");
                return EXIT_OK;
         }
         
         switch (ci->type) {	// Non parametrized commands
         
         	case ID_EXIT:               
             		logger("INFO", "Exit key detected");
                    	finishAll();                        
			return EXIT_ABORT;
                 
         	case ID_TMPDISCONN:
                	logger("INFO", "Temporary disconnect key detected");
                 	finishAll();
                 	return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
                                  
             	case ID_REMOTEOFF:
               		logger("INFO", "Switch remote off");
               		remoteOn = 0;
                	return EXIT_OK;
                        
            	case ID_TIMERCMD:   
                	logger("INFO", "Command: Timer");
                	if (serverMode != SERVER_NO) {
                		return processTimerCmd(ci);
                	}     
                	return EXIT_OK;

            	case ID_CANCELTIMER:   
                	logger("INFO", "Command: CancelTimer");
                	if (serverMode != SERVER_NO) {
                 		cancelTimer(ci->descr);
                	}  
                	return EXIT_OK;

            	case ID_SEND:
                	logger("INFO", "Command: Send");
                	if (serverMode != SERVER_NO) {
                        	ret = processSendCmd(ci);
                        	if (ret != EXIT_OK) {
                         		logger("DEBUG","Fails in processSendCmd()");
                        	}
                	}    
                	return EXIT_OK;
                        
                default:
                	break;        
        }
        
	switch (ci->type) {		// Parametrized commands, will parametrize in handler
                 
            	case ID_EXECSET: 
               		logger("INFO", "Command: ExecAndSet");
                	if (serverMode != SERVER_NO) { 
                		execCmdAndGetResults(ID_SET, ci, NULL, p);
               			logger("DEBUG", "Command: ExecAndSet FINISHED");
                        }
                	return EXIT_OK;
                 
            	case ID_EXECSEND:  
                	logger("INFO", "Command: ExecAndSend");
                	if (serverMode != SERVER_NO) {
               			execCmdAndGetResults(ID_SEND, ci, NULL, p);
                        }
                	return EXIT_OK;

            	case ID_MACRO:   
                	logger("INFO", "Command: Macro");
                 	ret = macroCmd(ci->descr,p);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
               		return EXIT_OK;
                 
                default:
                	break;        
        }

        if (ci->type == ID_EXEC) {	// Parametrized command, handler will use already parametrized data 
        	logger("INFO", "Command: Exec");
                
         	char* sExec = substParams(ci->exec, p);
        	if (sExec == NULL) {
        		logger("DEBUG", "Empty command after substParams()");
        		return EXIT_NOK;
        	}

                execSimpleCmd(sExec, 0);
                free(sExec);
                
                return EXIT_OK;
        }
        
        char* sDescr = substParams(ci->descr, p);
        if (sDescr == NULL) {
        	logger("DEBUG", "Empty command after substParams()");
                return EXIT_NOK;
        }
                       
	switch (ci->type) {		// Parametrized commands, handler will use already parametrized data 
                 
             	case ID_SETMODE:
                	logger("INFO", "Command: SetMode");
                         
                	int i = 0;    
                     	type_key* k = findItem(EVT_EXITMODE, &i, NULL);
                    	if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on exit from mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                                 	return ret;
                                 }
                      	} else {
                         	logger("DEBUG", "Leave mode");
                     	}
                      	setMode(sDescr);
                 
                     	i = 0; 
                     	k = findItem(EVT_ENTRMODE, &i, NULL);
                      	if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on enter to mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        		free(sDescr);
                                 	return ret;
                                 }
                   	} else {
                         	logger("DEBUG", "Entered to mode");
                       	}
                 	break;

             	case ID_SENDCKPD:
                	logger("INFO", "Command: SendCKPD");
                 	if (serverMode == SERVER_NO) {
                         
                        	ret = sendSeq(sDescr);
                        	if (ret != EXIT_OK) {
                         		logger("DEBUG","Fails in sendSeq()");
                         	}
                         	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        		free(sDescr);
                         		return ret;
                         	}
                	}    
                 	break;
                 
            	case ID_SET:          
                	logger("INFO", "Command: Set");
                	if (serverMode != SERVER_NO && sDescr != NULL && 
                            strncmp(sDescr,SET_UPLD,strlen(SET_UPLD)) != 0) {	// do not allow Set(upload,...);
                            
                        	strcpy(tmptext, CMD_SET);
                        	strcat(tmptext, "(");
                                
                 		//sprintf(tmp, "ID_SET >%s<", sDescr);
                 		//logger("DEBUG", tmp);
                 
                        	// Change \n in input to real '\n'
                        	char *p1 = sDescr;
                        	char *p2 = sDescr;
                        	while (p2 != NULL) {
                                	p2 = strstr(p1, "\\n");
                         
                                	if (p2 == NULL) {
                                 		strncat(tmptext, p1, p2-p1);
                                 	} else {
                                        	strncat(tmptext, p1, p2-p1);
                                        	strcat(tmptext, "\n");
                               		}
                                	p1 = p2 + 2;
                         	}
                        	strcat(tmptext, ");");
                 
                        	ret = writeSocketConn(tmptext);
                        	if (ret != EXIT_OK) {
                         		logger("DEBUG","Fails in writeSocketConn() (Set)");
                         	}
                                
                                /*if (strncmp(sDescr,SET_TEXT,strlen(SET_TEXT)) == 0 &&	// Send "final" replay
                                    (strncmp(sDescr+strlen(SET_TEXT)+1,"add",3) == 0 || 
                                     strncmp(sDescr+strlen(SET_TEXT)+1,"replace",7) == 0)) {
                                     
                                	logger("DEBUG", "Send final replay for Set(text,...)");	
                                        
                        		strcpy(tmptext, CMD_SET);
                        		strcat(tmptext, "(");
                        		strcat(tmptext, SET_TEXT);
                                        strcat(tmptext, ",add,SAME,<<);");
                         		ret = writeSocketConn(tmptext);
                        		if (ret != EXIT_OK) {
                         			logger("DEBUG","Fails in writeSocketConn() (Set)");
                         		}
                               }*/
            		}    
             		break;
                 
           	case ID_LOAD:   
               		logger("INFO", "Command: Load");
                 	ret = loadCmds(sDescr);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	free(sDescr);
                        	return ret;
                        }
                 	break;

           	case ID_INCLUDE:   
               		logger("INFO", "Command: Include");
                 	ret = includeCmd(sDescr);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	free(sDescr);
                        	return ret;
                        }
                 	break;
                 
		default:  
                	logger("INFO", "Unknown command");                 
                	break;
        }
	free(sDescr);
        
        return EXIT_OK;
}

int processCommands(cmdItem* cmdList,cmdParams* p) 
{
	while (cmdList) {
        	if (cmdList->descr == NULL) {
                        sprintf(tmp, "Process next command NULL parameters");
                } else {
                        sprintf(tmp, "Process next command %s",cmdList->descr);
                }
                logger("DEBUG", tmp);
                    
                int ret = processOneCommand(cmdList, p);
                if (ret != EXIT_OK) {
                	sprintf(tmp, "Return code from processOneCommand() is %d",ret);
                        logger("DEBUG", tmp);
                         
                        if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	logger("DEBUG", "Abort or disconnect after processOneCommand()");
                        
                                return ret;
                        }
                }
                cmdList = cmdList->next;
        }
        return EXIT_OK;
}

//
// Top-level command processing function
//
int handleCmdByKey(type_key* k, cmdParams* p)
{
	if (k == NULL) {
        	logger("DEBUG", "handleCmdByKey() NULL input");
                return EXIT_OK;
        }
        
        sprintf(tmp, "handleCmdByKey() >%s<",k->key);
        logger("DEBUG", tmp);
        
        if (p != NULL) {
	        sprintf(tmp,"handleCmdByKey params >%s,%s< ", p->index,p->value);
               	logger("DEBUG",tmp);
	}

        cmdItem* cmdList = getCommand(k);
        if (cmdList == NULL) {
        	logger("DEBUG", "No appropriate command was found by getCommand()");
        } else {
        	sprintf(tmp, "ready to execute command");
                logger("DEBUG", tmp);
        
                int ret = processCommands(cmdList, p);
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                	repeatNow    = 0;
                        repeatCmdPtr = NULL;
                        return ret;
                }
        
                // In AT mode: User had pressed some key ... 
                // Send ToMainMenu sequence of CKPD's to show main menu again
                if (serverMode == SERVER_NO) {
                	if (sendSeq(getToMainMenu()) != EXIT_OK) {
                        	logger("DEBUG","Fails in sendSeq()");
                        }
                }
        }   

        return EXIT_OK;     
}

//
// Functions related to timer commands
//

static int processTimerCmd(cmdItem* ci)
{
	if (ci->exec == NULL || ci->tdata == NULL) {
        	logger("DEBUG", "processTimerCmd(): nothing to exec or timer data absent");
                return EXIT_NOK;
        }
        
        timerCmd* tc = NULL;
	if ((tc = createTimer(ci)) == NULL) {
        	logger("DEBUG", "processTimerCmd(): timer already exists or input was incorrect");                
                return EXIT_NOK;
        }

        execCmdAndGetResults(ID_SET, ci, tc, NULL);

        return EXIT_OK;
}

timerCmd* findTimer(char *id, timerCmd **prev) 
{
	if (id == NULL) {
        	return NULL;
        }

	timerCmd* ptr = timers;

        while (ptr != NULL && strncmp(ptr->timer->tdata->id,id,MTEXTLEN) != 0) {
        	if (prev != NULL) {
        		*prev = ptr;
                }
                ptr  = ptr->next;
        }

        return ptr;
}

static timerCmd* createTimer(cmdItem *ci)
{
	if (ci->tdata->id[0] == '\0' || ci->exec == NULL) {
        	logger("DEBUG", "createTimer(): wrong input");                
                return NULL;
        }

        if (findTimer(ci->tdata->id, NULL) != NULL) {
        	logger("DEBUG", "createTimer(): timer already exists");                
                return NULL;
        } 
        timerCmd *tm = calloc(1,sizeof(timerCmd));

        //tm->type      = type;
        tm->ticks     = 0;
        tm->times     = 1;
        tm->timer     = ci;
        tm->lastValue = NULL;

        // Insert in head
        if (timers == NULL) { // Insert first
        	tm->next = NULL;
        } else {
        	tm->next = timers;
        }
        timers = tm;

        return tm;
}

int cancelTimer(char *id) 
{
	sprintf(tmp, "cancelTimer() >%s<", id);
        logger("DEBUG", tmp);

        timerCmd *prev = NULL;
        timerCmd *tm = findTimer(id, (timerCmd **) &prev);
        if (tm == NULL) {
            logger("DEBUG", "Cant find timer to cancel!");
            return EXIT_NOK;
        }    
        if (tm == timers) { //remove in head
        	timers = tm->next;
        } else {
        	if (prev != NULL) {
                	prev->next = tm->next;
                } else {
                	logger("ERROR", "Previous item absent for non-first element");
                        return EXIT_NOK;
                }
        }       
         
	tm->timer = NULL;
        free(tm->lastValue); tm->lastValue = NULL;
        free(tm);

        return EXIT_OK;
}

void freeTimers()
{
	logger("DEBUG", "freeTimers()");
        timerCmd *tm;

        while (timers != NULL) {
        	tm     = timers;
                timers = timers->next;
                
        	tm->timer = NULL;
                
                free(tm->lastValue); tm->lastValue = NULL;
                free(tm);
        }
        return;
}

int verifyTimers(int ticksInSec) 
{
	timerCmd *tm   = timers;

        while (tm != NULL) {
        	
                if (tm->ticks >= (tm->timer->tdata->timeout * ticksInSec)) {
                	sprintf(tmp, "verifyTimers(): it is time to execute >%s<", tm->timer->exec);
                        logger("DEBUG",tmp);

                        execCmdAndGetResults(ID_SET, tm->timer, tm, NULL);
                
                        tm->ticks = 0;
                
                        if (tm->timer->tdata->times > 0) {
                        	tm->times++;
                                if (tm->times >= tm->timer->tdata->times) {    // Cancel this timer 
                                	sprintf(tmp, "verifyTimers(): timer %s self-canceled", tm->timer->tdata->id);
                                        logger("DEBUG", tmp);
                                
                                        // Cancel this timer
                                        timerCmd *tmc = tm;
                                        tm = tm->next;

                                        cancelTimer(tmc->timer->tdata->id);
                                
                                        continue;
                                }
                        }
                } else {
                	tm->ticks++;
                }
                tm   = tm->next;
        }
	
        return EXIT_OK;
}

static void updateTimerLastData (timerCmd* timer, char *data)
{
	if (timer->lastValue != NULL) {
        	free(timer->lastValue);
                timer->lastValue = NULL;
        }
        timer->lastValue = calloc(strlen(data) + 1, 1);
        strcpy(timer->lastValue,data);
}
