/*
 * anyRemote
 * Copyright (C) 2006
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */


#ifndef _PARSE_H_
#define _PARSE_H_ 1

#include "common.h"

/* Constants */
#define MAXARGLEN   	256
#define MAXNAMELEN   	512
#define MAXCMDLEN   	1024

#define DEFAULT_DEVICE  "/dev/rfcomm0"

// Some AT stuff
// \r will be added before send
#define DEF_AT_CMER_ON_DEFAULT   "AT+CMER=3,2,0,0,0" 	// Default, valid for SE and Siemens S55 phones
#define DEF_AT_CMER_OFF_DEFAULT  "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SE        "AT+CMER=3,2,0,0,0" 	// SE, same as default
#define DEF_AT_CMER_OFF_SE       "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_MOTOROLA  "AT+CMER=3,1,0,0,0" 	// Motorola phones	
#define DEF_AT_CMER_OFF_MOTOROLA "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SAGEM     "AT+CMER=2,1,0,0,0" 	// Sagem phones
#define DEF_AT_CMER_OFF_SAGEM    "AT+CMER=0,0,0,0,0" 
#define DEF_AT_CMER_ON_SIEMENS   "AT+CMER=3,2,0,0,0" 	// Siemens phones, same as default
#define DEF_AT_CMER_OFF_SIEMENS  "AT+CMER=0,0,0,0,0" 

#define DEF_AT_CGMI     "AT+CGMI" 
#define DEF_AT_CLCC     "AT+CLCC"		// ? will be added later (will try both variants)
#define DEF_AT_CSCS     "AT+CSCS="
#define DEF_AT_CKPD 	"AT+CKPD="
#define DEF_AT_CLIP 	"AT+CLIP=1"
#define DEF_AT_CMEC 	"AT+CMEC=2"		// used by Siemens to made AT+CKPD works
#define DEF_CKEV 	"+CKEV:"
#define DEF_CLCC 	"+CLCC:"
#define DEF_RING 	"RING"			// It could be RING or +CRING (depends on AT+CRC)
#define DEF_MSG 	"Msg:"			// Used by Java Client
#define DEF_CLDEBUG 	"Msg:_DEBUG_("	        // Used by Java Client

#define CMER_ON  	1
#define CMER_OFF  	2

#define SAME 	    	99
#define PAUSE_STR   	"PAUSE"

#define EVT_CONNECT     "(Connect)"
#define EVT_DISCONNECT  "(Disconnect)"
#define EVT_ENDCALL     "(EndCall)"
#define EVT_INCALL     	"(InCall)"
#define EVT_INIT     	"(Init)"
#define EVT_ENTRMODE	"(EnterMode)"
#define EVT_EXITMODE    "(ExitMode)"

#define ID_UNKNOWN    	0
#define ID_EXIT    	1
#define ID_TMPDISCONN   2
#define ID_REMOTEON	3
#define ID_REMOTEOFF    4
#define ID_EXEC         5
#define ID_SENDCKPD     6
#define ID_SETMODE      7
#define ID_SET          8
#define ID_EXECSET      9
#define ID_TIMERCMD     10
#define ID_SEND         11
#define ID_EXECSEND     12
#define ID_CANCELTIMER  13
#define ID_MACRO        14
#define ID_LOAD         15
#define ID_INCLUDE      16

#define CMD_EXIT    	        "Exit"
#define CMD_TMPDISCONN	        "TempDisconnect"
#define CMD_REMOTEON	        "RemoteOn"
#define CMD_REMOTEOFF	        "RemoteOff"
#define CMD_EXEC                "Exec"
#define CMD_SENDCKPD            "SendCKPD"                  // AT mode only
#define CMD_SETMODE	        "SetMode"
#define CMD_SET                 "Set"                       // Server mode only
#define CMD_EXECSET             "ExecAndSet"                // Server mode only
#define CMD_TIMER               "Timer"                     // Server mode only
#define CMD_SEND                "Send"                      // Server mode only, mainly for Bemused-server-emulation
#define CMD_EXECSEND            "ExecAndSend"               // Server mode only, mainly for Bemused-server-emulation
#define CMD_CANCELTIMER         "CancelTimer"               // Server mode only
#define CMD_MACRO               "Macro"            
#define CMD_LOAD                "Load"            
#define CMD_INCLUDE             "Include"            

#define SET_BG         "bg"
#define SET_PARAM      "parameter"
#define SET_EFIELD     "editfield"
#define SET_FG         "fg"
#define SET_FMGR       "filemanager"
#define SET_FONT       "font"
#define SET_FSCREEN    "fullscreen"
#define SET_ICONS      "icons"
#define SET_LIST       "list"
#define SET_MENU       "menu"
#define SET_REPAINT    "repaint"
#define SET_SKIN       "skin"
#define SET_STATUS     "status"
#define SET_TEXT       "text"
#define SET_TITLE      "title"
#define SET_UPLD       "upload"
#define SET_VIBRATE    "vibrate"
#define SET_VOLUME     "volume"

#define STR_MOTOROLA	"Motorola"
#define STR_SE   	"Sony Ericsson"
#define STR_SAGEM	"SAGEM"
#define STR_SIEMENS   	"SIEMENS"

#define MODEL_DEFAULT	0
#define MODEL_SE	1
#define MODEL_MOTOROLA	2
#define MODEL_SAGEM	3
#define MODEL_SIEMENS	4

#define FLAG_EXACT	-1 // All non-negative values could be indexes in the list
#define FLAG_MULTIKEY	-2
#define FLAG_PARAMETR	 1
#define FLAG_UNDEF	-3

typedef struct Td
{
	char       id[MAXARGLEN];
	int        timeout;
        int	   times;
} timerData;

typedef struct Ci
{
	int        type;
        char       *descr;
        char       *exec;
        timerData  *tdata;
        struct Ci  *next;
} cmdItem;

typedef struct Tk
{
	char        *key;
        cmdItem     *cmd;
        struct Tk   *next;
} type_key;

typedef struct Ta
{
	char       key   [MAXARGLEN];
        char       alias [MAXARGLEN];
        struct Ta  *next;
} type_alias;

typedef struct Mt
{
	char        name[MTEXTLEN];
        type_key*   keys;
        struct Mt   *next;
} mode;

typedef struct Cp
{
	char       index[4];
        char       value[MAXARGLEN];
} cmdParams;


// Exported functions definitions

char* getAlarmInfo      (int which, cmdItem **cmds);
char* getAT_CMER     	(int what);
char* getCharset        (void);
cmdItem* getCommand	(type_key* item);
char* getDevice         (void);
char* getToMainMenu     (void);
type_key* findItem	(char *key, int *flag, cmdParams *param);
char* findAlias         (char *keyIn);
int   getAutoRepeat     (void);
int   getBaudrate       (void);
int   getLog            (void);
int   getDebug          (void);
int   getRetrySecs  	(void);
int   getUseScreen      (void);
int   getMaxTextSize    (void);
void  setMode           (char *modeName);
char* getMode           (void);
int   getModel          (void);
void  setModel          (char *m);
char* getCfgDir         (void);
int   getFrontEnd       (void);
int   init_cfg          (char *path);
void  freeCfg           (void);
void  freeCmds          (cmdItem* item);
int   autoConnect       (void);
int   isModifier       	(char *key);
int   parse_opts        (int argc, char *argv[]);
int   storeCmds         (cmdItem** cmdList, char *value);
void  printConf         (void);
int   load_cfg		(char *mfile, int isInit);
void  printKeys		(void);
void  printConf         (void);

#endif
